@@ -3,21 +3,24 @@
 __all__ = ['Savify']
 
 import time
-import os
-from uuid import uuid1
 from multiprocessing import cpu_count
 from multiprocessing.dummy import Pool as ThreadPool
+from pathlib import Path
+from urllib.error import URLError
 
 import validators
 import tldextract
 from youtube_dl import YoutubeDL
-from ffmpy import FFmpeg
+from ffmpy import FFmpeg, FFRuntimeError
+from requests.exceptions import ConnectionError
 
 from .utils import PathHolder, safe_path_string, check_env, check_ffmpeg, check_file, create_dir, clean
 from .types import *
 from .spotify import Spotify
 from .track import Track
-from .logger import Logger
+from .logger import Logger, LogLevel
+from .exceptions import FFmpegNotInstalledError, SpotifyApiCredentialsNotSetError, UrlNotSupportedError, \
+    YoutubeDlExtractionError, InternetConnectionError
 
 
 def _sort_dir(track, group):
@@ -37,33 +40,43 @@ def _progress(data):
     elif data['status'] == 'finished':
         pass
     elif data['status'] == 'error':
-        raise RuntimeError('youtube-dl download failed')
+        raise YoutubeDlExtractionError
 
 
 class Savify:
     def __init__(self, api_credentials=None, quality=Quality.BEST, download_format=Format.MP3,
-                 group=None, quiet: bool = False, path_holder: PathHolder = None, retry: int = 3,
-                 ydl_options: dict = {}, skip_cover_art: bool = False):
+                 group=None, log_level: int = LogLevel.QUIET, path_holder: PathHolder = None, retry: int = 3,
+                 ydl_options: dict = {}, skip_cover_art: bool = False, logger: Logger = None,
+                 ffmpeg_location: str = None):
 
         self.downloaded_cover_art = {}
         self.download_format = download_format
-        self.logger = Logger(quiet=quiet)
         self.path_holder = path_holder
         self.quality = quality
         self.group = group
-        self.quiet = quiet
         self.retry = retry
         self.ydl_options = ydl_options
         self.skip_cover_art = skip_cover_art
+        self.ffmpeg_location = ffmpeg_location
 
         if api_credentials is None:
             if not (check_env()):
-                raise RuntimeError('Spotify API credentials not setup.')
+                raise SpotifyApiCredentialsNotSetError
             else:
                 self.spotify = Spotify()
         else:
             self.spotify = Spotify(api_credentials=api_credentials)
 
+        if not check_ffmpeg() and self.ffmpeg_location is None:
+            raise FFmpegNotInstalledError
+
+        if logger is None:
+            self.logger = Logger(log_level=log_level)
+        else:
+            self.logger = logger
+
+        clean(self.path_holder.get_temp_dir())
+
     def _parse_query(self, query, query_type=Type.TRACK) -> list:
         result = []
 
@@ -72,7 +85,7 @@ class Savify:
             if domain == Platform.SPOTIFY:
                 result = self.spotify.link(query)
             else:
-                print('Invalid Spotify URL')
+                raise UrlNotSupportedError(query)
         else:
             if query_type == Type.TRACK:
                 result = self.spotify.search(query, query_type=Type.TRACK)
@@ -84,16 +97,16 @@ class Savify:
         return result
 
     def download(self, query, query_type=Type.TRACK) -> None:
-        if not (check_ffmpeg()):
-            print("FFmpeg must be installed to use Savify!")
-            return
-
-        queue = self._parse_query(query, query_type=query_type)
+        try:
+            queue = self._parse_query(query, query_type=query_type)
+        except ConnectionError or URLError:
+            raise InternetConnectionError
 
         if not (len(queue) > 0):
-            print('No tracks found using the given query.')
+            self.logger.info('Nothing found using the given query.')
             return
 
+        self.logger.info(f'Downloading {len(queue)} songs...')
         start_time = time.time()
         with ThreadPool(cpu_count())as pool:
             jobs = pool.map(self._download, queue)
@@ -103,39 +116,42 @@ class Savify:
                 if job['returncode'] != 0:
                     failed_jobs.append(job)
 
+        self.logger.info('Cleaning up...')
         clean(self.path_holder.get_temp_dir())
 
-        message = f'\nDownload Finished! \nCompleted {len(queue) - len(failed_jobs)}/{len(queue)}' \
-                  f' tracks in {time.time() - start_time:.4f}s '
+        message = f'Download Finished!\n\tCompleted {len(queue) - len(failed_jobs)}/{len(queue)}' \
+                  f' songs in {time.time() - start_time:.0f}s\n'
 
         if len(failed_jobs) > 0:
-            message += '\n\nFailed Tracks:\n'
+            message += '\n\tFailed Tracks:\n'
             for failed_job in failed_jobs:
-                message += f'\nTrack: {failed_job["track"].name}\nReason: {failed_job["error"]}\n'
+                message += f'\n\tSong:\t{str(failed_job["track"])}' \
+                           f'\n\tReason:\t{failed_job["error"]}\n'
 
-        print(message)
+        self.logger.info(message)
 
     def _download(self, track: Track) -> dict:
-        logger = Logger(quiet=self.quiet)
         status = {
             'track': track,
             'returncode': -1
         }
 
-        query = str(track) + ' (AUDIO)'
+        extractor = 'ytsearch'
+        query = f'{extractor}:{str(track)} audio'
         output = self.path_holder.get_download_dir() / f'{_sort_dir(track, self.group)}' / safe_path_string(
-            f'{track.artist_names[0]} - {track.name}.{self.download_format}')
+            f'{str(track)}.{self.download_format}')
 
-        output_temp = f'{str(self.path_holder.get_temp_dir())}/{str(uuid1())}.%(ext)s'
+        output_temp = f'{str(self.path_holder.get_temp_dir())}/{track.id}.%(ext)s'
 
         if check_file(output):
-            print('Song already downloaded. Skipping...')
+            self.logger.warning(f'{str(track)} -> is already downloaded. Skipping...')
             status['returncode'] = 0
             return status
 
         create_dir(output.parent)
 
         options = {
+            'quiet': self.logger.log_level <= LogLevel.QUIET,
             'format': 'bestaudio/best',
             'outtmpl': output_temp,
             'restrictfilenames': True,
@@ -143,8 +159,7 @@ class Savify:
             'nooverwrites': True,
             'noplaylist': True,
             'prefer_ffmpeg': True,
-            'default_search': 'ytsearch',
-            'logger': logger,
+            'logger': self.logger,
             'progress_hooks': [_progress],
 
             'postprocessors': [{
@@ -172,6 +187,9 @@ class Savify:
             options['postprocessor_args'].append('-codec:a')
             options['postprocessor_args'].append('libmp3lame')
 
+        if self.ffmpeg_location is not None:
+            options['ffmpeg_location'] = self.ffmpeg_location
+
         attempt = 0
         downloaded = False
 
@@ -181,69 +199,75 @@ class Savify:
             try:
                 with YoutubeDL(options) as ydl:
                     ydl.download([query])
-                    downloaded = True
-
-            except RuntimeError:
+                    if check_file(Path(output_temp)):
+                        downloaded = True
+            except YoutubeDlExtractionError as ex:
                 if attempt > self.retry:
                     status['returncode'] = 1
-                    status['error'] = "Failed to download track."
-                    print(logger.log)
+                    status['error'] = "Failed to download song."
+                    self.logger.error(ex.message)
                     return status
 
+        from shutil import move, Error as ShutilError
+
         if self.download_format != Format.MP3 or self.skip_cover_art:
             try:
-                import shutil
-                shutil.move(output_temp, output)
-            except Exception:
+                move(output_temp, output)
+            except ShutilError:
                 status['returncode'] = 1
-                status['error'] = "Unknown."
+                status['error'] = 'Filesystem error.'
+                self.logger.error('Failed to move temp file!')
                 return status
 
             status['returncode'] = 0
+            self.logger.info(f'Downloaded -> {str(track)}')
             return status
 
-        from ffmpy import FFRuntimeError
-
         attempt = 0
         added_artwork = False
 
         while not added_artwork:
             attempt += 1
 
+            cover_art_name = f'{track.album_name} - {track.artist_names[0]}'
+
+            if cover_art_name in self.downloaded_cover_art:
+                cover_art = self.downloaded_cover_art[cover_art_name]
+            else:
+                cover_art = self.path_holder.download_file(track.cover_art_url, extension='jpg')
+                self.downloaded_cover_art[cover_art_name] = cover_art
+
+            ffmpeg = FFmpeg(executable=self.ffmpeg_location,
+            inputs = {str(output_temp): None, str(cover_art): None, },
+                     outputs = {
+                str(output): '-loglevel quiet -hide_banner -y -map 0:0 -map 1:0 -c copy -id3v2_version 3 '
+                             '-metadata:s:v title="Album cover" -metadata:s:v comment="Cover (front)" '
+                # '-af "silenceremove=start_periods=1:start_duration=1:start_threshold=-60dB:'
+                # 'detection=peak,aformat=dblp,areverse,silenceremove=start_periods=1:'
+                # 'start_duration=1:start_threshold=-60dB:'
+                # 'detection=peak,aformat=dblp,areverse"'
+                }
+            )
+
             try:
-                cover_art_name = f'{track.album_name} - {track.artist_names[0]}'
-
-                if cover_art_name in self.downloaded_cover_art:
-                    cover_art = self.downloaded_cover_art[cover_art_name]
-                else:
-                    cover_art = self.path_holder.download_file(track.cover_art_url, extension='jpg')
-                    self.downloaded_cover_art[cover_art_name] = cover_art
-
-                ffmpeg = FFmpeg(
-                    inputs={str(output_temp): None, str(cover_art): None, },
-                    outputs={str(output): '-loglevel quiet -hide_banner -y -map 0:0 -map 1:0 -c copy -id3v2_version 3 '
-                                     '-metadata:s:v title="Album cover" -metadata:s:v comment="Cover (front)" '
-                             # '-af "silenceremove=start_periods=1:start_duration=1:start_threshold=-60dB:'
-                             # 'detection=peak,aformat=dblp,areverse,silenceremove=start_periods=1:'
-                             # 'start_duration=1:start_threshold=-60dB:'
-                             # 'detection=peak,aformat=dblp,areverse"'
-                             }
-                )
-
                 ffmpeg.run()
-
                 added_artwork = True
-
             except FFRuntimeError:
                 if attempt > self.retry:
                     try:
-                        import shutil
-                        shutil.move(output_temp, output)
+                        move(output_temp, output)
                         added_artwork = True
-                    except Exception:
+                    except ShutilError:
                         status['returncode'] = 1
-                        status['error'] = "Unknown."
+                        status['error'] = 'Filesystem error.'
+                        self.logger.error('Failed to move temp file!')
                         return status
 
         status['returncode'] = 0
+        try:
+            from os import remove
+            remove(output_temp)
+        except OSError:
+            pass
+        self.logger.info(f'Downloaded -> {str(track)}')
         return status
